from enum import Enum
from typing import Callable, List
from pika.exceptions import ChannelClosedByBroker
import sys

from whitson_tool_helper.logger import LOGGER


class MessagingSystem(Enum):
    pubsub = "pubsub"
    rabbitmq = "rabbitmq"


class MessagingSystemNotSupportedError(Exception):
    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return f"Messaging system {self.name} not supported at the moment"


class Consumer:
    def __init__(self, messaging_system: MessagingSystem, process_function: Callable):
        if messaging_system == MessagingSystem.pubsub:
            from whitson_tool_helper.messaging import pubsub

            self._consumer = pubsub.PubsubConsumer(process_function)
        elif messaging_system == MessagingSystem.rabbitmq:
            from whitson_tool_helper.messaging import rabbitmq

            self._consumer = rabbitmq.RabbitMQConsumer(process_function)
        else:
            raise MessagingSystemNotSupportedError(messaging_system)

        self._messaging_system = messaging_system

    def work(self):
        while True:
            LOGGER.info("Connecting to server...")
            try:
                self._consumer.work()
            except KeyboardInterrupt:
                LOGGER.info("Keyboard interrupt!")
                LOGGER.info("Consumer shut down!")
                sys.exit()
            except ChannelClosedByBroker as e:
                LOGGER.error("-- RABBITMQ ACCESS REFUSED --")
                LOGGER.error(e)
                sys.exit()
            except Exception as e:
                LOGGER.error("-- UNKNOWN ERROR ENCOUNTERED --")
                LOGGER.error(e)
            finally:
                if self._messaging_system == MessagingSystem.rabbitmq:
                    self._consumer.connection.close()
                LOGGER.warning("Restarting worker")


class Publisher:
    def __init__(self, messaging_system: MessagingSystem, exchange: str = None):
        if messaging_system == MessagingSystem.pubsub:
            from whitson_tool_helper.messaging import pubsub

            self._consumer = pubsub.PubsubPublisher()
        elif messaging_system == MessagingSystem.rabbitmq:
            from whitson_tool_helper.messaging import rabbitmq

            self._consumer = rabbitmq.RabbitMQPublisher(exchange)
        else:
            raise MessagingSystemNotSupportedError(messaging_system)

    def publish(self, topic: str, payload: dict, meta_data: dict = {}, **kwargs: dict):
        self._consumer.publish(topic, payload, meta_data, **kwargs)

    def publish_many(self, topic: str, payloads: List[dict], **kwargs: dict):
        self._consumer.publish_many(topic, payloads, **kwargs)



def apply_timeout(timeout_in_seconds: int):
    def decorator(func):
        def wrapper(*args, **kwargs):
            import signal
            from whitson_tool_helper import CalculationTimeoutError

            def handler(signum, frame):
                raise CalculationTimeoutError

            signal.signal(signal.SIGALRM, handler)
            signal.alarm(timeout_in_seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result

        return wrapper

    return decorator



import pika
import os
from typing import List
import json

from whitson_tool_helper.messaging.helper import check_environment_variables
from whitson_tool_helper.logger import LOGGER


def get_rabbitmq_params():
    check_environment_variables(
        [
            "RABBITMQ_USER",
            "RABBITMQ_PASSWORD",
            "RABBITMQ_HOST",
            "RABBITMQ_PORT",
        ],
    )
    credentials = pika.PlainCredentials(
        os.environ["RABBITMQ_USER"], os.environ["RABBITMQ_PASSWORD"]
    )
    return pika.ConnectionParameters(
        host=os.environ["RABBITMQ_HOST"],
        port=os.environ["RABBITMQ_PORT"],
        credentials=credentials,
    )


class RabbitMQConsumer:
    def __init__(self, process_function):
        self.process_function = process_function

        check_environment_variables(["MESSAGING_SUBSCRIPTION"])
        parameters = get_rabbitmq_params()
        self.connection = pika.BlockingConnection(parameters)
        self.channel = self.connection.channel()
        subscription = os.getenv("MESSAGING_SUBSCRIPTION")
        self.channel.queue_declare(
            queue=subscription,
            durable=True,
            arguments={"x-max-priority": 100, "x-queue-type": "classic"},
        )
        exchange = "engines"
        if subscription.endswith("-calculated"):
            exchange = "clients"

        self.channel.queue_bind(
            queue=subscription, exchange=exchange, routing_key=subscription
        )

        self.channel.basic_qos(prefetch_count=1)

        def callback(ch, method, properties, body):
            if not body or body.decode("utf-8") == "":
                ch.basic_ack(delivery_tag=method.delivery_tag)
                LOGGER.info("Recieved faulty message")
                LOGGER.info("Listening to", method.routing_key, "...")
                return

            msg = json.loads(body.decode("utf-8"))
            process_function(data=msg["data"], meta_data=msg["meta_data"])
            LOGGER.info("  ---  SUCCESS  ---  ")
            ch.basic_ack(delivery_tag=method.delivery_tag)
            LOGGER.info(f"Listening to {method.routing_key} ...")

        self.channel.basic_consume(queue=subscription, on_message_callback=callback)

    def work(self):
        self.channel.start_consuming()


class RabbitMQPublisher:
    def __init__(self, exchange):
        self.parameters = get_rabbitmq_params()
        self.exchange = exchange

    def publish(
        self, topic: str, payload: dict, meta_data: dict = {}, priority: int = 50
    ):
        connection = pika.BlockingConnection(self.parameters)
        channel = connection.channel()

        message = {"data": payload, "meta_data": meta_data}

        channel.basic_publish(
            exchange=self.exchange,
            routing_key=topic,
            body=json.dumps(message).encode("utf-8"),
            properties=pika.BasicProperties(
                delivery_mode=pika.spec.PERSISTENT_DELIVERY_MODE,
                priority=priority,
            ),
        )

        connection.close()

    def publish_many(self, topic, payloads: List[dict], priority: int = 50):
        connection = pika.BlockingConnection(self.parameters)
        channel = connection.channel()

        for message in payloads:
            channel.basic_publish(
                exchange="engines",
                routing_key=topic,
                body=json.dumps(message).encode("utf-8"),
                properties=pika.BasicProperties(
                    delivery_mode=pika.spec.PERSISTENT_DELIVERY_MODE,
                    priority=priority,
                ),
            )

        connection.close()



from typing import List
import os


def check_environment_variables(env_vars_to_check: List[str]):
    missing = []
    for var in env_vars_to_check:
        if os.getenv(var) is None:
            missing.append(var)
    if len(missing) > 0:
        from whitson_tool_helper import MissingEnvironmentVariableError

        raise MissingEnvironmentVariableError(missing)



from typing import List


class WhitsonException(Exception):
    pass


class CalculationTimeoutError(WhitsonException):
    def __str__(self) -> str:
        return "Calculation timed out. Please contact the support team."


class MissingCredentialsError(WhitsonException):
    def __init__(self, credential_type: str) -> None:
        self.credential_type = credential_type

    def __str__(self) -> str:
        return f"Credentials of {self.credential_type} are missing."


class MissingEnvironmentVariableError(WhitsonException):
    def __init__(self, vars: List[str]):
        self.vars = vars

    def __str__(self) -> str:
        return "Missing environment variables: " + ", ".join(self.vars)



import os
from google.cloud import pubsub_v1
import google.api_core.exceptions
import json
import traceback
from typing import List

from whitson_tool_helper.messaging.helper import check_environment_variables
from whitson_tool_helper.logger import LOGGER


class PubsubConsumer:
    def __init__(self, process_func, to_dict=False):
        check_environment_variables(
            [
                "GOOGLE_PROJECT",
                "GOOGLE_APPLICATION_CREDENTIALS",
            ]
        )
        self.decode_json = to_dict
        self.process_func = process_func
        self.subscription = os.environ["MESSAGING_SUBSCRIPTION"]

        self.google_project_name = os.environ["GOOGLE_PROJECT"]
        self.subscriber = pubsub_v1.SubscriberClient()

    @property
    def subscription_path(self):
        if self.subscriber:
            return self.subscriber.subscription_path(
                self.google_project_name, self.subscription
            )
        else:
            raise Exception("Pubsub subscriber not yet instantiated")

    def work(self):
        while True:
            LOGGER.info(f"Pulling subscription {self.subscription_path}")
            try:
                response = self.subscriber.pull(
                    subscription=self.subscription_path, max_messages=1
                )
            except google.api_core.exceptions.DeadlineExceeded as e:
                LOGGER.debug(e)
                continue
            except Exception as e:
                LOGGER.error(e)
                raise
            for received_message in response.received_messages:
                msg = received_message.message
                LOGGER.info(
                    f"Received PubSub message {msg.message_id}",
                    {"pubsub_msg_id": msg.message_id},
                )
                data = json.loads(msg.data.decode("utf8"))
                meta_data = msg.attributes
                try:
                    self.process_func(data, meta_data=meta_data)
                except Exception as e:
                    LOGGER.error(e)
                    LOGGER.error(traceback.format_exc())
                    LOGGER.warning(f"ACKING UNPROCESSED MESSAGE: {msg.message_id}")
                else:
                    LOGGER.info(f"ACKING MESSAGE: {msg.message_id}")
                self.subscriber.acknowledge(
                    subscription=self.subscription_path,
                    ack_ids=[received_message.ack_id],
                )


class PubsubPublisher:
    def __init__(self):
        check_environment_variables(["GOOGLE_PROJECT"])
        self.publisher = pubsub_v1.PublisherClient()
        self.project_id = os.environ["GOOGLE_PROJECT"]
        self.futures = set()

    @property
    def topic_name(self):
        return f"projects/{self.project_id}/topics/{self.topic}"

    def callback(future: pubsub_v1.publisher.futures.Future) -> None:
        future.result()

    def publish(
        self,
        topic: str,
        payload: dict,
        meta_data: dict = {},
        encode_json=True,
        **kwargs,
    ):
        LOGGER.debug("Publishing single to pubsub")
        self.topic = topic
        payload = json.dumps(payload).encode("utf8") if encode_json else payload
        future = self.publisher.publish(self.topic_name, payload, **meta_data)
        future.result()

    def publish_many(self, topic, payloads: List[dict]):
        LOGGER.debug("Publishing many to pubsub")
        self.topic = topic

        for payload in payloads:
            data = json.dumps(payload.get("data")).encode("utf-8")
            meta_data = payload.get("meta_data") or {}
            future = self.publisher.publish(self.topic_name, data, **meta_data)
            future.result()
